import numpy as np

from geoana.em.base import BaseMagneticDipole
from geoana.em.tdem.base import BaseTDEM

from geoana.em.tdem.simple_functions import magnetic_field_vertical_magnetic_dipole, magnetic_field_time_deriv_magnetic_dipole


class VerticalMagneticDipoleHalfSpace(BaseMagneticDipole, BaseTDEM):
    """Transient of a vertical magnetic dipole in a half space.

    Only valid for source and receivers at the surface. The surface is assumed
    to be at z=0.

    Waveform is assumed to be the step off.
    """

    def magnetic_field(self, xy):
        """Magnetic field due to a magnetic dipole over a half space

        The analytic expression is only valid for a source and receiver at the
        surface of the earth.

        Parameters
        ----------
        xy : (n_locations, 2) numpy.ndarray
            receiver locations

        Returns
        -------
        numpy.ndarray
            magnetic field for each xy location
        """
        dxy = xy - self.location
        h = magnetic_field_vertical_magnetic_dipole(
            np.r_[self.time], dxy[:, :2], self.sigma, self.mu, self.moment
        )
        return h[0]  # because time was a 1 element array

    def magnetic_flux_density(self, xy):
        """Magnetic flux due to a step off magnetic dipole over a half space

        The analytic expression is only valid for a source and receiver at the
        surface of the earth.

        Parameters
        ----------
        xy : (n_locations, 2) numpy.ndarray
            receiver locations

        Returns
        -------
        numpy.ndarray
            magnetic flux for each xy location
        """
        return self.mu * self.magnetic_field(xy)

    def magnetic_field_time_derivative(self, xy):
        """Magnetic flux time derivative due to a step off magnetic dipole over a half space

        The analytic expression is only valid for a source and receiver at the
        surface of the earth.

        Parameters
        ----------
        xy : (n_locations, 2) numpy.ndarray
            receiver locations

        Returns
        -------
        numpy.ndarray
            Magnetic flux time derivative for each xy location
        """

        dxy = xy - self.location
        dh_dt = magnetic_field_time_deriv_magnetic_dipole(
            np.r_[self.time], dxy[:, :2], self.sigma, self.mu, self.moment
        )
        return dh_dt[0]

    def magnetic_flux_time_derivative(self, xy):
        """Magnetic flux due to a step off magnetic dipole over a half space

        The analytic expression is only valid for a source and receiver at the
        surface of the earth.

        Parameters
        ----------
        xy : (n_locations, 2) numpy.ndarray
            receiver locations

        Returns
        -------
        numpy.ndarray
            magnetic flux for each xy location
        """
        return self.mu * self.magnetic_field_time_derivative(xy)



"""
oksar3

Program to calcuate forward models of interferograms, strain tensor, etc.
from Okada subroutine.

Heritage:
    - originally fringes.c written by Barry Parsons
    - updated to oksar                                 tjw
    - oksar_strain: added strain tensor calculation    tjw
    - oksar3:       added new line of sight calculator tjw feb 2003
    - Modified into Python by RowanCockett, 3point Science Aug 2014


"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import properties
import vectormath as vmath
import utm
import matplotlib.pyplot as plt


class EarthquakeInterferogram(properties.HasProperties):
    title = properties.String(
        'name of the earthquake',
        required=True
    )

    description = properties.String(
        'description of the event',
        required=False
    )

    location = properties.Vector2(
        'interferogram location (bottom N, left E)',
        required=True
    )

    location_UTM_zone = properties.Integer(
        'UTM zone',
        required=True
    )

    shape = properties.Array(
        'number of pixels in the interferogram',
        shape=(2,),
        dtype=int,
        required=True
    )

    pixel_size = properties.Array(
        'Size of each pixel (northing, easting)',
        shape=(2,),
        dtype=float,
        required=True
    )

    data = properties.Array(
        'Processed interferogram data (unwrapped)',
        dtype=float,
        required=True
    )

    ref = properties.Vector2(
        'interferogram reference',
        required=True
    )

    ref_incidence = properties.Float(
        'Incidence angle',
        required=True
    )

    scaling = properties.Float(
        'Scaling of the interferogram',
        default=1.0
    )

    satellite_name = properties.String('Name of the satelite.')

    satellite_fringe_interval = properties.Float(
        'Fringe interval',
        default=0.028333
    )

    satellite_azimuth = properties.Float(
        'satellite_azimuth',
        required=True
    )

    satellite_altitude = properties.Float(
        'satellite_altitude',
        required=True
    )

    local_rigidity = properties.Float(
        'Local rigidity',
        default=3e10
    )

    local_earth_radius = properties.Float(
        'Earth radius',
        default=6371000.
    )

    date1 = properties.DateTime(
        'date1',
        required=True
    )

    date2 = properties.DateTime(
        'date2',
        required=True
    )

    processed_by = properties.String(
        'processed_by',
        required=True
    )

    processed_date = properties.DateTime(
        'processed_date',
        required=True
    )

    copyright = properties.String(
        'copyright',
        required=True
    )

    data_source = properties.String(
        'data_source',
        required=True
    )

    event_date = properties.DateTime('Date of the earthquake')
    event_gcmt_id = properties.String('GCMT ID')
    event_name = properties.String('Earthquake name')
    event_country = properties.String('Earthquake country')

    def _get_plot_data(self):

        vectorNx = (
            np.r_[
                0,
                np.cumsum(
                    (self.pixel_size[0],) * self.shape[0]
                )
            ] + self.location[0]
        )
        vectorNy = (
            np.r_[
                0,
                np.cumsum(
                    (self.pixel_size[1],) * self.shape[1]
                )
            ] + self.location[1]
        ) - self.pixel_size[1] * self.shape[1]

        data = self.data.copy()
        data = np.flipud(data.reshape(self.shape, order='F').T)
        data[data == 0] = np.nan
        data *= self.scaling

        return vectorNx, vectorNy, data

    def plot_interferogram(self, wrap=True, ax=None):

        self.assert_valid

        if ax is None:
            plt.figure()
            ax = plt.subplot(111)

        vectorNx, vectorNy, data = self._get_plot_data()

        if wrap:
            cmap = plt.cm.hsv
            data = data % self.satellite_fringe_interval
            vmin, vmax = 0.0, self.satellite_fringe_interval
        else:
            cmap = plt.cm.jet
            vmin = np.nanmin(data)
            vmax = np.nanmax(data)

        out = ax.pcolormesh(
            vectorNx,
            vectorNy,
            np.ma.masked_where(np.isnan(data), data),
            vmin=vmin,
            vmax=vmax,
            cmap=cmap
        )

        ax.set_title(self.title)
        ax.axis('equal')
        ax.set_xlabel('Easting, m (UTM Zone {})'.format(
            self.location_UTM_zone
        ))
        ax.set_ylabel('Northing, m')

        cb = plt.colorbar(out, ax=ax)
        cb.set_label('Displacement, m')

        return out

    def plot_mask(self, ax=None, opacity=0.2):

        if ax is None:
            plt.figure()
            ax = plt.subplot(111)

        vectorNx, vectorNy, data = self._get_plot_data()

        from matplotlib import colors
        cmap = colors.ListedColormap([(1, 1, 1, opacity)])

        out = ax.pcolormesh(
            vectorNx,
            vectorNy,
            np.ma.masked_where(~np.isnan(data), data),
            cmap=cmap
        )

        ax.set_title(self.title)
        ax.axis('equal')
        ax.set_xlabel('Easting, m (UTM Zone {})'.format(
            self.location_UTM_zone
        ))
        ax.set_ylabel('Northing, m')

        return out

    def get_LOS_vector(self, locations):
        """
        calculate beta - the angle at earth center between reference point
        and satellite nadir
        """
        if not isinstance(locations, list):
            locations = [locations]

        utmZone = self.location_UTM_zone
        refPoint = vmath.Vector3(self.ref.x, self.ref.y, 0)
        satAltitude = self.satellite_altitude
        satAzimuth = self.satellite_azimuth
        satIncidence = self.ref_incidence
        earthRadius = self.local_earth_radius

        DEG2RAD = np.pi / 180.
        alpha = satIncidence * DEG2RAD
        beta = (earthRadius / (satAltitude + earthRadius)) * np.sin(alpha)
        beta = alpha - np.arcsin(beta)
        beta = beta / DEG2RAD

        # calculate angular separation of (x,y) from satellite track passing
        # through (origx, origy) with azimuth satAzimuth

        # Long lat **NOT** lat long
        origy, origx = utm.to_latlon(
            refPoint.x, refPoint.y, np.abs(utmZone), northern=utmZone > 0
        )

        xy = np.array([
            utm.to_latlon(u[0], u[1], np.abs(utmZone), northern=utmZone > 0)
            for u in locations
        ])
        y = xy[:, 0]
        x = xy[:, 1]

        angdist = self._ang_to_gc(x, y, origx, origy, satAzimuth)

        # calculate beta2, the angle at earth center between roaming point and
        # satellite nadir track, assuming right-looking satellite

        beta2 = beta - angdist
        beta2 = beta2 * DEG2RAD

        # calculate alpha2, the new incidence angle

        alpha2 = np.sin(beta2) / (
            np.cos(beta2) - (earthRadius / (earthRadius + satAltitude))
        )
        alpha2 = np.arctan(alpha2)
        alpha2 = alpha2 / DEG2RAD

        # calculate pointing vector

        satIncidence = 90 - alpha2
        satAzimuth = 360 - satAzimuth

        los_x = -np.cos(satAzimuth * DEG2RAD) * np.cos(satIncidence * DEG2RAD)
        los_y = -np.sin(satAzimuth * DEG2RAD) * np.cos(satIncidence * DEG2RAD)
        los_z = np.sin(satIncidence * DEG2RAD)

        return vmath.Vector3Array([los_x, los_y, los_z])

    @staticmethod
    def _ang_to_gc(x, y, origx, origy, satAzimuth):
        """
        Calculate angular distance to great circle passing through
        given point
        """

        Ngc = np.zeros(3)
        cartxy = np.zeros((len(x), 3))
        satAzimuth = np.deg2rad(satAzimuth)
        origx = np.deg2rad(origx)
        origy = np.deg2rad(origy)

        x = np.deg2rad(x)
        y = np.deg2rad(y)

        # 1. calc geocentric norm vec to great circle, Ngc = Rz*Ry*Rx*[0;1;0]
        #    where Rz = rotation of origx about geocentric z-axis
        #    where Ry = rotation of origy about geocentric y-axis
        #    where Rx = rotation of satAzimuth about geocentric x-axis
        #    and [0;1;0] is geocentric norm vec to N-S Great Circle through 0 0

        Ngc[0] = (
                    (np.sin(satAzimuth) * np.sin(origy) * np.cos(origx)) -
                    (np.cos(satAzimuth) * np.sin(origx))
                 )
        Ngc[1] = (
                    (np.sin(satAzimuth) * np.sin(origy) * np.sin(origx)) +
                    (np.cos(satAzimuth) * np.cos(origx))
                 )
        Ngc[2] = -np.sin(satAzimuth) * np.cos(origy)

        # 2. calculate unit vector geocentric coordinates for lon/lat
        #    position (x,y)

        cartxy[:, 0] = np.cos(x) * np.cos(y)
        cartxy[:, 1] = np.sin(x) * np.cos(y)
        cartxy[:, 2] = np.sin(y)

        # 3. Dot product between Ngc and cartxy gives angle 90 degrees
        #    bigger than what we want

        angdist = (
                        Ngc[0]*cartxy[:, 0] +
                        Ngc[1]*cartxy[:, 1] +
                        Ngc[2]*cartxy[:, 2]
                  )

        angdist = np.rad2deg(np.arccos(angdist)) - 90

        return angdist


class Oksar(properties.HasProperties):

    beta = properties.Float('beta', default=3E10)
    mu = properties.Float('mu', default=3E10)

    strike = properties.Float('Strike', min=0, max=360)
    dip = properties.Float('Dip', default=45, min=0, max=90)
    rake = properties.Float('Rake', default=90, min=-180, max=180)
    slip = properties.Float('Slip', default=0.5, min=0)
    length = properties.Float('Fault length', default=10000., min=0)
    center = properties.Vector2('Center of the fault plane.')
    depth_top = properties.Float('Top of fault', min=0)
    depth_bottom = properties.Float('Bottom of fault', default=10000, min=0)

    O = properties.Vector2(
        'Origin of the simulation domain', required=True
    )

    U = properties.Vector2(
        'U direction of the simulation domain', required=True
    )

    V = properties.Vector2(
        'V direction of the simulation domain', required=True
    )

    shape = properties.Array(
        'number of pixels in the simulation',
        shape=(2,),
        default=(300, 300),
        dtype=int,
        # required=True
    )

    @property
    def simulation_grid(self):

        self.assert_valid

        vec, shape = vmath.ouv2vec(
            vmath.Vector3(self.O[0], self.O[1], 0),
            vmath.Vector3(self.U[0], self.U[1], 0),
            vmath.Vector3(self.V[0], self.V[1], 0),
            self.shape
        )
        return vec

    @property
    def displacement_vector(self):

        self.assert_valid

        vec = self.simulation_grid
        x, y = vec.x, vec.y

        DEG2RAD = 0.017453292519943
        alpha = (self.beta + self.mu) / (self.beta + 2.0 * self.mu)

        #  Here we could loop over models

        flt_x = self.center[0]
        flt_y = self.center[1]
        strike = self.strike
        dip = self.dip
        rake = self.rake
        slip = self.slip
        length = self.length
        hmin = self.depth_top
        hmax = self.depth_bottom

        rrake = (rake+90.0)*DEG2RAD
        sindip = np.sin(dip*DEG2RAD)
        w = (hmax-hmin)/sindip
        ud = slip*np.cos(rrake)
        us = -slip*np.sin(rrake)
        halflen = length/2.0
        al2 = halflen
        al1 = -al2
        aw1 = hmin/sindip
        aw2 = hmax/sindip

        if(hmin < 0.0):
            raise Exception('ERROR: Fault top above ground surface')

        if(hmin == 0.0):
            hmin = 0.00001

        sstrike = (strike+90.0)*DEG2RAD

        ct = np.cos(sstrike)
        st = np.sin(sstrike)

        X = ct * (-flt_x + x) - st * (-flt_y + y)
        Y = ct * (-flt_y + y) + st * (-flt_x + x)

        u = self._dc3d3(alpha, X, Y, -dip, al1, al2, aw1, aw2, us, ud)

        UX = ct*u.x + st*u.y
        UY = -st*u.x + ct*u.y
        UZ = u.z

        return vmath.Vector3(UX, UY, UZ)

    def _dc3d3(self, alpha, X, Y, dip, al1, al2, aw1, aw2, disl1, disl2):
        F0 = 0.0
        F1 = 1.0
        F2 = 2.0
        PI2 = 6.283185307179586
        EPS = 1.0E-6

        u = vmath.Vector3(F0, F0, F0)
        dub = vmath.Vector3(F0, F0, F0)

        #  %%dccon0 subroutine
        #  Calculates medium and fault dip constants
        c0_alp3 = (F1 - alpha) / alpha
        #  PI2/360
        pl8 = 0.017453292519943
        c0_sd = np.sin(dip*pl8)
        c0_cd = np.cos(dip*pl8)

        if(np.abs(c0_cd) < EPS):
            c0_cd = F0
            if(c0_sd > F0):
                c0_sd = F1

            if(c0_sd < F0):
                c0_sd = -F1

        c0_cdcd = c0_cd * c0_cd
        c0_sdcd = c0_sd * c0_cd

        #  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        p = Y * c0_cd
        q = Y * c0_sd

        jxi = ((X - al1) * (X - al2)) <= F0  # BOOLEAN
        jet = ((p - aw1) * (p - aw2)) <= F0  # BOOLEAN

        for k in [1., 2.]:
            et = 0.0
            if(k == 1):
                et = p-aw1
            else:
                et = p-aw2

            for j in [1., 2.]:
                xi = 0.0
                if(j == 1):
                    xi = X-al1
                else:
                    xi = X-al2

                # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                # %%dccon2 subroutine
                # % calculates station geometry constants for finite source

                dc_max = np.max(np.abs(np.c_[xi, et, q]))

                # dc_max = max(np.abs(xi),max(np.abs(et),np.abs(q)))

                xi[(np.abs(xi/dc_max) < EPS) | (np.abs(xi) < EPS)] = F0

                et[(np.abs(et/dc_max) < EPS) | (np.abs(et) < EPS)] = F0

                q[(np.abs(q/dc_max) < EPS) | (np.abs(q) < EPS)] = F0

                dc_xi = xi
                dc_et = et
                dc_q = q
                c2_r = np.sqrt(dc_xi*dc_xi + dc_et*dc_et + dc_q*dc_q)

                if np.any(c2_r == F0):
                    raise Exception('singularity error ???')

                c2_y = dc_et * c0_cd + dc_q * c0_sd
                c2_d = dc_et * c0_sd - dc_q * c0_cd
                c2_tt = np.arctan(dc_xi * dc_et / (dc_q * c2_r))
                c2_tt[dc_q == F0] = F0

                rxi = c2_r + dc_xi
                c2_x11 = F1/(c2_r*rxi)
                c2_x11[(dc_xi < F0) & (dc_q == F0) & (dc_et == F0)] = F0

                ret = c2_r + dc_et
                if np.any(ret < 1e-14):
                    raise Exception('dccon2 b %f %f %f %f' % (
                        ret, c2_r, dc_et, dc_q, dc_xi
                    ))

                c2_ale = np.log(ret)
                c2_y11 = F1/(c2_r*ret)

                ind = (dc_et < F0) & (dc_q == F0) & (dc_xi == F0)

                # if((c2_r-dc_et) < 1e-14):
                #     raise Exception('dccon2 a %f %f %f %f %f' % (
                #         c2_3-dc_et, c2_r, dc_et, dc_q, dc_xi)
                #     )

                c2_ale[ind] = -np.log(c2_r[ind]-dc_et[ind])
                c2_y11[ind] = F0

                # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

                if np.any(
                    (
                        (q == F0) &
                        (
                            ((jxi) & (et == F0)) |
                            ((jet) & (xi == F0))
                        )
                    ) | (c2_r == F0)
                ):
                    raise Exception('singular problems: 2')

                # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                # ub subroutine
                # part B of displacement and strain at depth due to buried
                # faults in semi-infinite medium

                rd = c2_r + c2_d
                if np.any(rd < 1e-14):
                    raise Exception('ub %f %f %f %f %f %f' % (
                        rd, c2_r, c2_d, xi, et, q
                    ))

                ai3 = 0.0
                ai4 = 0.0
                if(c0_cd != F0):
                    # xx replaces x in original subroutine
                    xx = np.sqrt(xi*xi+q*q)
                    ai4 = F1/c0_cdcd * (xi/rd*c0_sdcd + F2*np.arctan(
                        (
                            et*(xx+q*c0_cd) +
                            xx*(c2_r+xx)*c0_sd
                        ) / (xi*(c2_r+xx)*c0_cd)
                    ))
                    ai4[xi == F0] = F0

                    ai3 = (c2_y*c0_cd/rd - c2_ale + c0_sd*np.log(rd)) / c0_cdcd
                else:
                    rd2 = rd*rd
                    ai3 = (et/rd + c2_y*q/rd2 - c2_ale) / F2
                    ai4 = xi*c2_y/rd2/F2

                ai1 = -xi/rd*c0_cd - ai4*c0_sd
                ai2 = np.log(rd) + ai3*c0_sd
                qx = q*c2_x11
                qy = q*c2_y11

                # strike-slip contribution
                if(disl1 != 0.0):
                    du2x = - xi*qy - c2_tt - c0_alp3 * ai1 * c0_sd
                    du2y = - q/c2_r + c0_alp3*c2_y/rd*c0_sd
                    du2z = q*qy - c0_alp3*ai2*c0_sd
                    du2 = vmath.Vector3(du2x, du2y, du2z)
                    dub = du2 * (disl1 / PI2)
                else:
                    dub = vmath.Vector3()

                # dip-slip contribution
                if(disl2 != F0):
                    du2x = - q/c2_r + c0_alp3 * ai3 * c0_sdcd
                    du2y = - et*qx - c2_tt - c0_alp3 * xi / rd * c0_sdcd
                    du2z = q*qx + c0_alp3 * ai4 * c0_sdcd
                    du2 = vmath.Vector3(du2x, du2y, du2z)
                    dub = dub + (du2 * (disl2 / PI2))

                # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

                dux = dub.x
                duy = dub.y*c0_cd - dub.z*c0_sd
                duz = dub.y*c0_sd + dub.z*c0_cd
                du = vmath.Vector3(dux, duy, duz)
                if((j+k) != 3):
                    u = + du + u
                else:
                    u = - du + u

        return u

    def plot_displacement(self, eq=None, ax=None, wrap=True, mask_opacity=0.2):

        if eq is not None:
            assert isinstance(eq, EarthquakeInterferogram)

        if ax is None:
            plt.figure()
            ax = plt.subplot(111)

        vectorNx = (
            np.r_[
                0,
                np.cumsum(
                    (self.U[0]/self.shape[0],) * self.shape[0]
                )
            ] + self.O[0]
        )
        vectorNy = (
            np.r_[
                0,
                np.cumsum(
                    (self.V[1]/self.shape[1],) * self.shape[1]
                )
            ] + self.O[1]
        )

        DIR = self.displacement_vector
        grid = self.simulation_grid
        if eq is None:
            LOS = vmath.Vector3(0, 0, 1)
        else:
            LOS = eq.get_LOS_vector(grid)
        data = DIR.dot(LOS)
        data = np.flipud(data.reshape(self.shape, order='F').T)
        # data[data == 0] = np.nan
        # data *= self.scaling

        if wrap and eq is not None:
            cmap = plt.cm.hsv
            data = data % eq.satellite_fringe_interval
            vmin, vmax = 0.0, eq.satellite_fringe_interval
        else:
            cmap = plt.cm.jet
            vmin = np.nanmin(data)
            vmax = np.nanmax(data)

        out = ax.pcolormesh(
            vectorNx,
            vectorNy,
            np.ma.masked_where(np.isnan(data), data),
            vmin=vmin,
            vmax=vmax,
            cmap=cmap
        ),

        ax.axis('equal')
        ax.set_xlabel('Easting, m')
        ax.set_ylabel('Northing, m')

        cb = plt.colorbar(out[0], ax=ax)
        cb.set_label('Displacement, m')

        if eq is not None:
            mask = eq.plot_mask(ax=ax, opacity=mask_opacity)
            out = out[0], mask

        return out


def example():

    import requests
    dinar_file = requests.get(
        'https://storage.googleapis.com/simpeg/geoana/dinar.r4'
    )
    data = np.fromstring(dinar_file.content, np.float32)

    dinar = EarthquakeInterferogram(
        # uid='dinar',
        title='Dinar, Turkey',
        description=(
            'On October 1, 1995, a strong earthquake ruptured a section of '
            'the Dinar-Civril fault in SW Turkey. Around 30% of the buildings '
            'in the nearby town of Dinar were destroyed. 92 inhabitants were '
            'killed and over 200 injured.'
        ),
        event_country='Turkey',
        event_date='1995-09-30T18:00:00Z',
        event_gcmt_id='100195B',
        event_name='Dinar',
        copyright='ESA',
        data_source='ESA',
        date1='1995-08-12T18:00:00Z',
        date2='1995-12-31T17:00:00Z',
        processed_by='GarethFunning',
        processed_date='2003-01-20T17:00:00Z',
        ref_incidence=23,
        ref=[741140., 4230327.],
        scaling=0.0045040848895,
        local_earth_radius=6386232,
        local_rigidity=30000000000,
        location=[706216.0606, 4269238.9999],
        shape=(1024, 1024),
        location_UTM_zone=35,
        pixel_size=[80., 80.],
        satellite_altitude=788792,
        satellite_azimuth=192,
        satellite_fringe_interval=0.028333333,
        satellite_name='ERS',
        data=data
    )

    dinar_fwd = Oksar(
        beta=3E10,
        mu=3E10,
        strike=329.6,
        dip=50,
        rake=90,
        slip=0.5,
        length=11578.907244622129,
        center=[773728.2977967655, 4223586.816611591],
        depth_top=0,
        depth_bottom=15000,
        O=[706216.0606, 4187318.9999],
        U=[81920, 0],
        V=[0, 81920],
        shape=(300, 200)
    )

    return dinar, dinar_fwd



import numpy as np
from scipy.special import erf, ive
from scipy.constants import mu_0
from geoana.em.tdem.base import theta


def vertical_magnetic_field_horizontal_loop(
    t, sigma=1.0, mu=mu_0, radius=1.0, current=1.0, turns=1
):
    """Central vertical magnetic field due to step off of a horizontal loop

    Parameters
    ----------
    t : float, or numpy.ndarray
    sigma : float, optional
        conductivity
    mu : float, optional
        magnetic permeability
    radius : float, optional
        radius of the horizontal loop
    current : float, optional
        current of the horizontal loop
    turns : int, optional
        number of turns in the horizontal loop

    Returns
    -------
    hz : float, or numpy.ndarray
        The vertical magnetic field at the center of the loop.
        The shape will match the `t` input.

    Notes
    -----
    Equation 4.98 in Ward and Hohmann 1988

    .. math::
        h_z = \\frac{I}{2a}\\left[
            \\frac{3}{\\sqrt{\\pi} \\theta a}e^{-\\theta^2 a^2}
            + \\left(1 - \\frac{3}{2 \\theta^2 a^2}\\right)\\mathrm{erf}(\\theta a)
            \\right]

    Examples
    --------
    Reproducing part of Figure 4.8 from Ward and Hohmann 1988

    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from geoana.em.tdem import vertical_magnetic_field_horizontal_loop

    Calculate the field at the time given

    >>> times = np.logspace(-7, -1)
    >>> hz = vertical_magnetic_field_horizontal_loop(times, sigma=1E-2, radius=50)

    Match the vertical magnetic field plot

    >>> plt.loglog(times*1E3, hz)
    >>> plt.xlabel('time (ms)')
    >>> plt.ylabel('H$_z$ (A/m)')
    >>> plt.show()

    """
    theta = np.sqrt((sigma * mu_0) / (4 * t))
    ta = theta * radius
    eta = erf(ta)
    t1 = (3 / (np.sqrt(np.pi) * ta)) * np.exp(-(ta ** 2))
    t2 = (1 - (3 / (2 * ta ** 2))) * eta
    hz = (t1 + t2) / (2 * radius)
    return turns * current * hz


def vertical_magnetic_flux_horizontal_loop(
    t, sigma=1.0, mu=mu_0, radius=1.0, current=1.0, turns=1
):
    """Central vertical magnetic flux density due to step off of a horizontal loop

    Parameters
    ----------
    t : float, or numpy.ndarray
    sigma : float, optional
        conductivity
    mu : float, optional
        magnetic permeability
    radius : float, optional
        radius of the horizontal loop
    current : float, optional
        current of the horizontal loop
    turns : int, optional
        number of turns in the horizontal loop

    Returns
    -------
    bz : float, or numpy.ndarray
        The vertical magnetic field at the center of the loop.
        The shape will match the `t` input.

    See Also
    --------
    vertical_magnetic_field_horizontal_loop
    """
    return mu * vertical_magnetic_field_horizontal_loop(
        t, sigma=sigma, mu=mu, radius=radius, current=current, turns=turns
    )


def vertical_magnetic_field_time_deriv_horizontal_loop(
    t, sigma=1.0, mu=mu_0, radius=1.0, current=1.0, turns=1
):
    """Central vertical magnetic field time derivative due to step off horizontal loop

    Parameters
    ----------
    t : float, or numpy.ndarray
    sigma : float, optional
        conductivity
    mu : float, optional
        magnetic permeability
    radius : float, optional
        radius of the horizontal loop
    current : float, optional
        current of the horizontal loop
    turns : int, optional
        number of turns in the horizontal loop

    Returns
    -------
    dhz_dt : float, or numpy.ndarray
        The vertical magnetic field time derivative at the center of the loop.
        The shape will match the `t` input.

    Notes
    -----
    Matches equation 4.97 of Ward and Hohmann 1988.

    .. math::
        \\frac{\\partial h_z}{\\partial t} = -\\frac{I}{\\sigma a^3}\\left[
        3 \\mathrm{erf}(\\theta a)
        - \\frac{2}{\\sqrt{\\pi}}\\theta a (3 + 2 \\theta^2 a^2)e^{-\\theta^2 a^2}
        \\right]

    Examples
    --------
    Reproducing part of Figure 4.8 from Ward and Hohmann 1988

    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from geoana.em.tdem import vertical_magnetic_field_time_deriv_horizontal_loop

    Calculate the field at the time given

    >>> times = np.logspace(-7, -1)
    >>> dhz_dt = vertical_magnetic_field_time_deriv_horizontal_loop(times, sigma=1E-2, radius=50)

    Match the vertical magnetic field plot

    >>> plt.loglog(times*1E3, -dhz_dt, '--')
    >>> plt.xlabel('time (ms)')
    >>> plt.ylabel(r'$\\frac{\\partial h_z}{ \\partial t}$ (A/(m s)')
    >>> plt.show()
    """
    a = radius
    the = theta(t, sigma, mu)
    return -turns * current / (mu * sigma * a**3) * (
        3*erf(the * a) - 2/np.sqrt(np.pi) * the * a * (3 + 2 * the**2 * a**2) * np.exp(-the**2 * a**2)
    )


def vertical_magnetic_flux_time_deriv_horizontal_loop(
    t, sigma=1.0, mu=mu_0, radius=1.0, current=1.0, turns=1
):
    """Central vertical magnetic flux density time derivative due to step off horizontal loop

    Parameters
    ----------
    t : float, or numpy.ndarray
    sigma : float, optional
        conductivity
    mu : float, optional
        magnetic permeability
    radius : float, optional
        radius of the horizontal loop
    current : float, optional
        current of the horizontal loop
    turns : int, optional
        number of turns in the horizontal loop

    Returns
    -------
    dbz_dt : float, or numpy.ndarray
        The vertical magnetic flux time derivative at the center of the loop.
        The shape will match the `t` input.

    See Also
    --------
    vertical_magnetic_field_time_deriv_horizontal_loop

    Examples
    --------
    Reproducing part of Figure 4.8, scaled by magnetic suscpetibility, from Ward and
    Hohmann 1988.

    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from geoana.em.tdem import vertical_magnetic_flux_time_deriv_horizontal_loop

    Calculate the field at the time given

    >>> times = np.logspace(-7, -1)
    >>> dbz_dt = vertical_magnetic_flux_time_deriv_horizontal_loop(times, sigma=1E-2, radius=50)

    Match the vertical magnetic field plot

    >>> plt.loglog(times*1E3, -dbz_dt, '--')
    >>> plt.xlabel('time (ms)')
    >>> plt.ylabel(r'$\\frac{\\partial b_z}{ \\partial t}$ (T/s)')
    >>> plt.show()
    """
    a = radius
    the = theta(t, sigma, mu)
    return -turns * current / (sigma * a**3) * (
        3*erf(the * a) - 2/np.sqrt(np.pi) * the * a * (3 + 2 * the**2 * a**2) * np.exp(-the**2 * a**2)
    )


def magnetic_field_vertical_magnetic_dipole(
    t, xy, sigma=1.0, mu=mu_0, moment=1.0
):
    """Magnetic field due to step off vertical dipole at the surface

    Parameters
    ----------
    t : (n_t) numpy.ndarray
        times (s)
    xy : (n_locs, 2) numpy.ndarray
        surface field locations (m)
    sigma : float, optional
        conductivity
    mu : float, optional
        magnetic permeability
    moment : float, optional
        moment of the dipole

    Returns
    -------
    h : (n_t, n_locs, 3) numpy.ndarray
        The magnetic field at the observation locations and times.

    Notes
    -----
    Matches the negative of equation 4.69a of Ward and Hohmann 1988, for the vertical
    component (due to the difference in coordinate sign conventionn used here).

    .. math::
        h_z = -\\frac{m}{4 \\pi \\rho^2} \\left[
        \\left(\\frac{9}{2 \\theta^2 \\rho^2} - 1\\right)\\mathrm{erf}(\\theta \\rho)
        - \\frac{1}{\\sqrt{\\pi}}\\left(\\frac{9}{\\theta \\rho + 4 \\theta \\rho} \\right)
        e^{-\\theta^2\\rho^2}
        \\right]

    Also matches equation 4.72 for the horizontal components, which is again negative due
    to our coordinate convention.

    .. math::
        h_\\rho = \\frac{m \\theta^2}{2\\pi\\rho}
        e^{-\\theta^2\\rho^2/2}\\left[I_1\\left(\\frac{\\theta^2\\rho^2}{2}\\right)
         - I_2\\left(\\frac{\\theta^2\\rho^2}{2}\\right)\\right]

    Examples
    --------
    Reproducing part of Figure 4.4 and 4.5 from Ward and Hohmann 1988

    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from geoana.em.tdem import magnetic_field_vertical_magnetic_dipole

    Calculate the field at the time given, and 100 m along the x-axis,

    >>> times = np.logspace(-8, 0, 200)
    >>> xy = np.array([[100, 0, 0]])
    >>> h = magnetic_field_vertical_magnetic_dipole(times, xy, sigma=1E-2)

    Match the vertical magnetic field plot

    >>> plt.loglog(times*1E3, h[:,0, 2], c='C0', label='$h_z$')
    >>> plt.loglog(times*1E3, -h[:,0, 2], '--', c='C0')
    >>> plt.loglog(times*1E3, h[:,0, 0], c='C1', label='$h_x$')
    >>> plt.loglog(times*1E3, -h[:,0, 0], '--', c='C1')
    >>> plt.xlabel('time (ms)')
    >>> plt.ylabel('h (A/m)')
    >>> plt.legend()
    >>> plt.show()
    """
    r = np.linalg.norm(xy[:, :2], axis=-1)
    x = xy[:, 0]
    y = xy[:, 1]
    thr = theta(t, sigma, mu=mu)[:, None] * r #will be positive...

    h_z = 1.0 / r**3 * (
        (9 / (2 * thr**2) - 1) * erf(thr)
        - (9 / thr + 4 * thr) / np.sqrt(np.pi) * np.exp(-thr**2)
    )
    # positive here because z+ up

    # iv(1, arg) - iv(2, arg)
    # ive(1, arg) * np.exp(abs(arg)) - ive(2, arg) * np.exp(abs(arg))
    # (ive(1, arg) - ive(2, arg))*np.exp(abs(arg))
    h_r = 2 * thr**2 / r**3 * (
        ive(1, thr**2 / 2) - ive(2, thr**2 / 2)
    )
    # thetar is always positive so this above simplifies (more numerically stable)

    angle = np.arctan2(y, x)
    h_x = np.cos(angle) * h_r
    h_y = np.sin(angle) * h_r

    return moment / (4 * np.pi) * np.stack((h_x, h_y, h_z), axis=-1)


def magnetic_field_time_deriv_magnetic_dipole(
    t, xy, sigma=1.0, mu=mu_0, moment=1.0
):
    """Magnetic field time derivative due to step off vertical dipole at the surface

    Parameters
    ----------
    t : (n_t) numpy.ndarray
        times (s)
    xy : (n_locs, 2) numpy.ndarray
        surface field locations (m)
    sigma : float, optional
        conductivity
    mu : float, optional
        magnetic permeability
    moment : float, optional
        moment of the dipole

    Returns
    -------
    dh_dt : (n_t, n_locs, 3) numpy.ndarray
        The magnetic field at the observation locations and times.

    Notes
    -----
    Matches the negative of equation 4.70 of Ward and Hohmann 1988, for the vertical
    component (due to the difference in coordinate sign conventionn used here).

    .. math::
        \\frac{\\partial h_z}{\\partial t} = \\frac{m}{2 \\pi \\mu \\sigma \\rho^5}\\left[
        9\\mathrm{erf}(\\theta \\rho)
        - \\frac{2\\theta\\rho}{\\sqrt{\\pi}}\\left(
        9 + 6\\theta^2\\rho^2 + 4\\theta^4\\rho^4\\right)
        e^{-\\theta^2\\rho^2}
        \\right]

    Also matches equation 4.74 for the horizontal components

    .. math::
        \\frac{\\partial h_\\rho}{\\partial t} = -\\frac{m \\theta^2}{2 \\pi \\rho t}e^{-\\theta^2\\rho^2/2} \\left[
        (1 +\\theta^2 \\rho^2) I_0 \\left(
        \\frac{\\theta^2\\rho^2}{2}
        \\right) - \\left(
        2 + \\theta^2\\rho^2 + \\frac{4}{\\theta^2\\rho^2}\\right)I_1\\left(\\frac{\\theta^2\\rho^2}{2}\\right)
        \\right]

    Examples
    --------
    Reproducing the time derivate parts of Figure 4.4 and 4.5 from Ward and Hohmann 1988

    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from geoana.em.tdem import magnetic_field_time_deriv_magnetic_dipole

    Calculate the field at the time given, 100 m along the x-axis,

    >>> times = np.logspace(-6, 0, 200)
    >>> xy = np.array([[100, 0, 0]])
    >>> dh_dt = magnetic_field_time_deriv_magnetic_dipole(times, xy, sigma=1E-2)

    Match the vertical magnetic field plot

    >>> plt.loglog(times*1E3, dh_dt[:,0, 2], c='C0', label=r'$\\frac{\\partial h_z}{\\partial t}$')
    >>> plt.loglog(times*1E3, -dh_dt[:,0, 2], '--', c='C0')
    >>> plt.loglog(times*1E3, dh_dt[:,0, 0], c='C1', label=r'$\\frac{\\partial h_x}{\\partial t}$')
    >>> plt.loglog(times*1E3, -dh_dt[:,0, 0], '--', c='C1')
    >>> plt.xlabel('time (ms)')
    >>> plt.ylabel(r'$\\frac{\\partial h}{\\partial t}$ (A/(m s))')
    >>> plt.legend()
    >>> plt.show()
    """
    r = np.linalg.norm(xy[:, :2], axis=-1)
    x = xy[:, 0]
    y = xy[:, 1]
    tr = theta(t, sigma, mu)[:, None] * r

    dhz_dt = 1 / (r**3 * t[:, None]) * (
        9 / (2 * tr**2) * erf(tr)
        - (4 * tr**3 + 6 * tr + 9/tr)/np.sqrt(np.pi)*np.exp(-tr**2)
    )

    # iv(k, v) = ive(k, v) * exp(abs(arg))
    dhr_dt = - 2 * tr**2 / (r**3 * t[:, None]) * (
        (1 + tr**2) * ive(0, tr**2 / 2) -
        (2 + tr**2 + 4 / tr**2) * ive(1, tr**2 / 2)
    )
    angle = np.arctan2(y, x)
    dhx_dt = np.cos(angle) * dhr_dt
    dhy_dt = np.sin(angle) * dhr_dt
    return moment / (4 * np.pi) * np.stack((dhx_dt, dhy_dt, dhz_dt), axis=-1)


def magnetic_flux_vertical_magnetic_dipole(
    t, xy, sigma=1.0, mu=mu_0, moment=1.0
):
    """
    Magnetic flux density due to step off vertical dipole at the surface

    Parameters
    ----------
    t : (n_t) numpy.ndarray
        times (s)
    xy : (n_locs, 2) numpy.ndarray
        surface field locations (m)
    sigma : float, optional
        conductivity
    mu : float, optional
        magnetic permeability
    moment : float, optional
        moment of the dipole

    Returns
    -------
    b : (n_t, n_locs, 3) numpy.ndarray
        The magnetic flux at the observation locations and times.

    See Also
    --------
    magnetic_field_vertical_magnetic_dipole
    """
    return mu * magnetic_field_vertical_magnetic_dipole(
        t, xy, sigma=sigma, mu=mu, moment=moment
    )


def magnetic_flux_time_deriv_magnetic_dipole(
    t, xy, sigma=1.0, mu=mu_0, moment=1.0
):
    """
    Magnetic flux density time derivative due to step off vertical dipole at the surface

    Parameters
    ----------
    t : (n_t) numpy.ndarray
        times (s)
    xy : (n_locs, 2) numpy.ndarray
        surface field locations (m)
    sigma : float, optional
        conductivity
    mu : float, optional
        magnetic permeability
    moment : float, optional
        moment of the dipole

    Returns
    -------
    db_dt : (n_t, n_locs, 3) numpy.ndarray
        The magnetic flux at the observation locations and times.

    See Also
    --------
    magnetic_field_time_deriv_magnetic_dipole
    """
    return mu * magnetic_field_time_deriv_magnetic_dipole(
        t, xy, sigma=sigma, mu=mu, moment=moment
    )



from geoana.em.tdem.base import peak_time, diffusion_distance, theta, BaseTDEM

from geoana.em.tdem.wholespace import ElectricDipoleWholeSpace

from geoana.em.tdem.halfspace import VerticalMagneticDipoleHalfSpace

from geoana.em.tdem.simple_functions import (
    vertical_magnetic_field_horizontal_loop,
    vertical_magnetic_flux_horizontal_loop,
    vertical_magnetic_field_time_deriv_horizontal_loop,
    vertical_magnetic_flux_time_deriv_horizontal_loop,
    magnetic_field_vertical_magnetic_dipole,
    magnetic_field_time_deriv_magnetic_dipole,
    magnetic_flux_vertical_magnetic_dipole,
    magnetic_flux_time_deriv_magnetic_dipole,
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from .utils import mkvc


def cylindrical_2_cartesian(grid, vec=None):
    """
    Take a grid or vector (if provided)
    defined in cylindrical coordinates :math:`(r, \\theta, z)` and
    transform it to cartesian coordinates, :math:`(x, y, z)`.

    **Required**

    :param numpy.ndarray grid: grid in cylindrical coordinates
                               :math:`(r, \\theta, z)`

    **Optional**

    :param numpy.ndarray vec: (optional) vector defined in cylindrical
                              coordinates

    **Returns**

    :return: grid or vector (if provided) in cartesian coordinates
             :math:`(x, y, z)`
    :rtype: numpy.ndarray

    """
    grid = np.atleast_2d(grid)

    if vec is None:
        return np.hstack([
            mkvc(grid[:, 0]*np.cos(grid[:, 1]), 2),
            mkvc(grid[:, 0]*np.sin(grid[:, 1]), 2),
            mkvc(grid[:, 2], 2)
        ])

    if len(vec.shape) == 1 or vec.shape[1] == 1:
        vec = vec.reshape(grid.shape, order='F')

    x = vec[:, 0] * np.cos(grid[:, 1]) - vec[:, 1] * np.sin(grid[:, 1])
    y = vec[:, 0] * np.sin(grid[:, 1]) + vec[:, 1] * np.cos(grid[:, 1])

    newvec = [x, y]
    if grid.shape[1] == 3:
        z = vec[:, 2]
        newvec += [z]

    return np.vstack(newvec).T


def cartesian_2_cylindrical(grid, vec=None):
    """
    Takes a grid or vector (if provided)
    defined in cartesian coordinates :math:`(x, y, z)` and
    transform it to cylindrical coordinates, :math:`(r, \\theta, z)`.

    **Required**

    :param numpy.ndarray grid: grid in cartesian coordinates
                               :math:`(x, y, z)`

    **Optional**

    :param numpy.ndarray vec: (optional) vector defined in cartesian
                              coordinates

    **Returns**

    :return: grid or vector (if provided) in cylindrical coordinates
             :math:`(r, \\theta, z)`
    :rtype: numpy.ndarray
    """

    grid = np.atleast_2d(grid)

    if vec is None:
        return np.hstack([
            mkvc(np.sqrt(grid[:, 0]**2 + grid[:, 1]**2), 2),
            mkvc(np.arctan2(grid[:, 1], grid[:, 0]), 2),
            mkvc(grid[:, 2], 2)
        ])

    if len(vec.shape) == 1 or vec.shape[1] == 1:
        vec = vec.reshape(grid.shape, order='F')

    theta = np.arctan2(grid[:, 1], grid[:, 0])

    return np.hstack([
        mkvc(np.cos(theta)*vec[:, 0] + np.sin(theta)*vec[:, 1], 2),
        mkvc(-np.sin(theta)*vec[:, 0] + np.cos(theta)*vec[:, 1], 2),
        mkvc(vec[:, 2], 2)
    ])


def spherical_2_cartesian(grid, vec=None):
    """
    Take a grid or vector (if provided)
    defined in spherical coordinates :math:`(r, \\theta, \\phi)` and
    transform it to cartesian coordinates, :math:`(x, y, z)`.

    **Required**

    :param numpy.ndarray grid: grid in spherical coordinates
                             :math:`(r, \\theta, \\phi)`

    **Optional**

    :param numpy.ndarray vec: (optional) vector defined in spherical
                              coordinates

    **Returns**

    :return: grid or vector (if provided) in cartesian coordinates
             :math:`(x, y, z)`
    :rtype: numpy.ndarray

    """
    grid = np.atleast_2d(grid)

    if vec is None:
        return np.hstack([
            mkvc(grid[:, 0] * np.sin(grid[:, 2]) * np.cos(grid[:, 1]), 2),
            mkvc(grid[:, 0] * np.sin(grid[:, 2]) * np.sin(grid[:, 1]), 2),
            mkvc(grid[:, 0] * np.cos(grid[:, 2]), 2)
        ])

    if len(vec.shape) == 1 or vec.shape[1] == 1:
        vec = vec.reshape(grid.shape, order='F')

    x = (
        vec[:, 0] * np.sin(grid[:, 2]) * np.cos(grid[:, 1]) +
        vec[:, 2] * np.cos(grid[:, 2]) * np.cos(grid[:, 1]) -
        vec[:, 1] * np.sin(grid[:, 1])
    )
    y = (
        vec[:, 0] * np.sin(grid[:, 2]) * np.sin(grid[:, 1]) +
        vec[:, 2] * np.cos(grid[:, 2]) * np.sin(grid[:, 1]) -
        vec[:, 1] * np.cos(grid[:, 1])
    )
    z = (
        vec[:, 0] * np.cos(grid[:, 2]) -
        vec[:, 2] * np.sin(grid[:, 2])
    )

    newvec = [x, y, z]

    return np.vstack(newvec).T


def cartesian_2_spherical(grid, vec=None):
    """
    Takes a grid or vector (if provided)
    defined in cartesian coordinates :math:`(x, y, z)` and
    transform it to spherical coordinates, :math:`(r, \\theta, \\phi)`.

    **Required**

    :param numpy.ndarray grid: grid in cartesian coordinates
                               :math:`(x, y, z)`

    **Optional**

    :param numpy.ndarray vec: (optional) vector defined in cartesian
                              coordinates

    **Returns**

    :return: grid or vector (if provided) in spherical coordinates
             :math:`(r, \\theta, \\phi)`
    :rtype: numpy.ndarray
    """

    grid = np.atleast_2d(grid)

    if vec is None:
        return np.hstack([
            mkvc(np.sqrt(grid[:, 0]**2 + grid[:, 1]**2 + grid[:, 2]**2), 2),
            mkvc(np.arctan2(grid[:, 1], grid[:, 0]), 2),
            mkvc(
                np.arctan2(np.sqrt(grid[:, 0]**2 + grid[:, 1]**2), grid[:, 2]),
                2
            ),
        ])

    if len(vec.shape) == 1 or vec.shape[1] == 1:
        vec = vec.reshape(grid.shape, order='F')

    theta = np.arctan2(grid[:, 1], grid[:, 0])
    phi = np.arctan2(np.sqrt(grid[:, 0]**2 + grid[:, 1]**2), grid[:, 2])

    r = (
        vec[:, 0] * np.sin(phi) * np.cos(theta) +
        vec[:, 1] * np.sin(phi) * np.sin(theta) +
        vec[:, 2] * np.cos(phi)
    )

    theta = - vec[:, 0] * np.sin(theta) + vec[:, 1] * np.cos(theta)

    phi = (
        vec[:, 0] * np.cos(phi) * np.cos(theta) +
        vec[:, 1] * np.cos(phi) * np.sin(theta) -
        vec[:, 2] * np.sin(phi)
    )

    newvec = [r, theta, phi]

    return np.vstack(newvec).T


def vector_magnitude(v):
    """
    Amplitude of a vector, v.

    **Required**

    :param numpy.ndarray v: vector array

    **Returns**

    :returns: magnitude of a vector (n, 1)
    :rtype: numpy.ndarray
    """

    v = np.atleast_2d(v)

    return np.sqrt((v**2).sum(axis=1))


def vector_distance(xyz, origin=np.r_[0., 0., 0.]):
    """
    Vector distance of a grid, xyz from an origin origin.

    **Required**

    :param numpy.ndarray xyz: grid (npoints x 3)

    **Optional**

    :param numpy.ndarray origin: origin (default: [0., 0., 0.])

    **Returns**

    :returns: vector distance from a grid of points from the origin
              (npoints x 3)
    :rtype: numpy.ndarray
    """
    assert(xyz.shape[1] == 3), (
        "the xyz grid should be npoints by 3, the shape provided is {}".format(
            xyz.shape
        )
    )

    if len(origin) != 3:
        raise Exception(
            "the origin must be length 3, the length provided is {}".format(
                len(origin)
            )
        )


    dx = xyz[:, 0] - origin[0]
    dy = xyz[:, 1] - origin[1]
    dz = xyz[:, 2] - origin[2]

    return np.c_[dx, dy, dz]


def distance(xyz, origin=np.r_[0., 0., 0.]):
    """
    Radial distance from an grid of points to the origin

    **Required**

    :param numpy.ndarray xyz: grid (npoints x 3)

    **Optional**

    :param numpy.ndarray origin: origin (default: [0., 0., 0.])

    **Returns**

    :returns: distance between each point and the origin (npoints x 1)
    :rtype: numpy.ndarray
    """
    dxyz = vector_distance(xyz, origin)
    return vector_magnitude(dxyz)


def vector_dot(xyz, vector):
    """
    Take a dot product between an array of vectors, xyz and a vector [x, y, z]

    **Required**

    :param numpy.ndarray xyz: grid (npoints x 3)
    :param numpy.ndarray vector: vector (1 x 3)

    **Returns**

    :returns: dot product between the grid and the (1 x 3) vector, returns an
              (npoints x 1) array
    :rtype: numpy.ndarray
    """
    if len(vector) != 3:
        raise Exception(
            "vector should be length 3, the provided length is {}".format(
                len(vector)
            )
        )
    return vector[0]*xyz[:, 0] + vector[1]*xyz[:, 1] + vector[2]*xyz[:, 2]


def repeat_scalar(scalar, dim=3):
    """
    Repeat a spatially distributed scalar value dim times to simplify
    multiplication with a vector.

    **Required**

    :param numpy.ndarray scalar: (n x 1) array of scalars

    **Optional**

    :param int dim: dimension of the second axis for the output (default = 3)

    **Returns**

    :returns: (n x dim) array of the repeated vector
    :rtype: numpy.ndarray
    """
    assert len(scalar) in scalar.shape, (
        "input must be a scalar. The shape you provided is {}".format(
            scalar.shape
        )
    )

    return np.kron(np.ones((1, dim)), np.atleast_2d(scalar).T)

def rotation_matrix_from_normals(v0, v1, tol=1e-20):
    """
    Performs the minimum number of rotations to define a rotation from the
    direction indicated by the vector n0 to the direction indicated by n1.
    The axis of rotation is n0 x n1
    https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula

    :param numpy.ndarray v0: vector of length 3
    :param numpy.ndarray v1: vector of length 3
    :param float tol: tolerance. If the norm of the cross product between the
                        two vectors is below this, no rotation is performed
                        default = 1e-20
    :rtype: numpy.ndarray
    :return: 3 x 3 rotation matrix which rotates the frame so that n0 is
             aligned with n1
    """

    v0 = mkvc(v0)
    v1 = mkvc(v1)

    # ensure both n0, n1 are vectors of length 1
    assert len(v0) == 3, "Length of n0 should be 3"
    assert len(v1) == 3, "Length of n1 should be 3"

    # ensure both are true normals
    n0 = v0*1./np.linalg.norm(v0)
    n1 = v1*1./np.linalg.norm(v1)

    n0dotn1 = n0.dot(n1)

    # define the rotation axis, which is the cross product of the two vectors
    rotAx = np.cross(n0, n1)

    if np.linalg.norm(rotAx) < tol:
        return np.eye(3, dtype=float)

    rotAx *= 1./np.linalg.norm(rotAx)

    cosT = n0dotn1/(np.linalg.norm(n0)*np.linalg.norm(n1))
    sinT = np.sqrt(1.-n0dotn1**2)

    ux = np.array(
        [
            [0., -rotAx[2], rotAx[1]],
            [rotAx[2], 0., -rotAx[0]],
            [-rotAx[1], rotAx[0], 0.]
        ], dtype=float
    )

    return np.eye(3, dtype=float) + sinT*ux + (1.-cosT)*(ux.dot(ux))


def rotate_points_from_normals(xyz, n0, n1, x0=np.r_[0., 0., 0.]):
    """
    rotates a grid so that the vector n0 is aligned with the vector n1

    **Required**

    :param numpy.ndarray xyz:
    :param numpy.ndarray n0: vector of length 3, should have norm 1
    :param numpy.ndarray n1: vector of length 3, should have norm 1

    **Optional**

    :param numpy.ndarray x0: vector of length 3, point about which we perform the
                           rotation

    **Returns**

    :rtype: numpy.ndarray
    :return: (3x3) rotation matrix which rotates the frame so that n0 is
             aligned with n1

    """

    R = rotation_matrix_from_normals(n0, n1)

    if xyz.shape[1] != 3:
        raise AssertionError("Grid xyz should be 3 wide")

    return (xyz - x0)@R.T + x0



from geoana.em.fdem.base import (
    omega, wavenumber, skin_depth, sigma_hat, BaseFDEM
)

from geoana.em.fdem.wholespace import (
    ElectricDipoleWholeSpace, MagneticDipoleWholeSpace
)

from geoana.em.fdem.halfspace import MagneticDipoleHalfSpace

from geoana.em.fdem.layered import MagneticDipoleLayeredHalfSpace

from geoana.em.fdem.simple_functions import (
    vertical_magnetic_field_horizontal_loop, vertical_magnetic_flux_horizontal_loop
)



from scipy.special import erf
import numpy as np
from geoana.em.tdem.base import BaseTDEM
from geoana.spatial import repeat_scalar
from geoana.em.base import BaseElectricDipole

###############################################################################
#                                                                             #
#                                  Classes                                    #
#                                                                             #
###############################################################################


class ElectricDipoleWholeSpace(BaseElectricDipole, BaseTDEM):
    """
    Harmonic electric dipole in a whole space. The source is
    (c.f. Ward and Hohmann, 1988 page 173). The source current
    density for a dipole located at :math:`\mathbf{r}_s` with orientation
    :math:`\mathbf{\hat{u}}`

    .. math::

        \mathbf{J}(\mathbf{r}) = I ds \delta(\mathbf{r}
        - \mathbf{r}_s)\mathbf{\hat{u}}

    """

    def electric_field(self, xyz):
        """
        Electric field from an electric dipole

        .. math::

            \mathbf{E} = \\frac{1}{\hat{\sigma}} \\nabla \\nabla \cdot \mathbf{A}
            - i \omega \mu \mathbf{A}

        """
        dxyz = self.vector_distance(xyz)
        r = self.distance(xyz)
        r = repeat_scalar(r)
        thetar = self.theta * r
        root_pi = np.sqrt(np.pi)

        front = (
            (self.current * self.length) / (4 * np.pi * self.sigma * r**3)
        )

        symmetric_term = (
            (
                - (
                    4/root_pi * thetar ** 3 + 6/root_pi * thetar
                ) * np.exp(-thetar**2) +
                3 * erf(thetar)
            ) * (
                repeat_scalar(self.dot_orientation(dxyz)) * dxyz / r**2
            )
        )

        oriented_term = (
            (
                4./root_pi * thetar**3 + 2./root_pi * thetar
            ) * np.exp(-thetar**2) -
            erf(thetar)
        ) * np.kron(self.orientation, np.ones((dxyz.shape[0], 1)))

        return front * (symmetric_term + oriented_term)

    def current_density(self, xyz):
        """
        Current density due to a harmonic electric dipole
        """
        return self.sigma * self.electric_field(xyz)

    def magnetic_field(self, xyz):
        """
        Magnetic field from an electric dipole
        """
        dxyz = self.vector_distance(xyz)
        r = self.distance(dxyz)
        r = repeat_scalar(r)
        thetar = self.theta * r

        front = (
            self.current * self.length / (4 * np.pi * r**2) * (
                2 / np.sqrt(np.pi) * thetar * np.exp(-thetar**2) + erf(thetar)
            )
        )

        return - front * self.cross_orientation(xyz) / r

    def magnetic_field_time_deriv(self, xyz):
        """
        Time derivative of the magnetic field,
        :math:`\\frac{\partial \mathbf{h}}{\partial t}`
        """

        dxyz = self.vector_distance(xyz)
        r = self.distance(dxyz)
        r = repeat_scalar(r)

        front = (
            self.current * self.length * self.theta**3 * r /
            (2 * np.sqrt(np.pi)**3 * self.time)
        )

        return - front * self.cross_orientation(xyz) / r

    def magnetic_flux_density(self, xyz):
        """
        Magnetic flux density from an electric dipole
        """

        return self.mu * self.magnetic_field(xyz)

    def magnetic_flux_density_time_deriv(self, xyz):
        """
        Time derivative of the magnetic flux density from an electric dipole
        """

        return self.mu * self.magnetic_field_time_deriv(xyz)



def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration("em", parent_package, top_path)

    config.add_subpackage("fdem")
    config.add_subpackage("tdem")

    return config



import numpy as np
from scipy.constants import mu_0
import properties
from scipy.special import iv, kv
from geoana.em.base import BaseMagneticDipole
from geoana.em.fdem.base import BaseFDEM


class MagneticDipoleHalfSpace(BaseMagneticDipole, BaseFDEM):
    """Harmonic magnetic dipole in a half space.

    Only valid for source and receivers at the surface. The surface is assumed
    to be at z=0.
    """

    frequency = properties.Array(
        "Source frequency (Hz)",
        shape=('*', ),
        dtype=float
    )

    @properties.validator("location")
    def _check_source_height(self, change):
        if change["value"][2] != 0.0:
            raise ValueError("Source must be at the surface of the earth (z=0)")

    def magnetic_field(self, xy, field="secondary"):
        """Magnetic field due to a magnetic dipole over a half space

        The analytic expression is only valid for a source and receiver at the
        surface of the earth. For arbitrary source and receiver locations above
        the earth, use the layered solution.

        Parameters
        ----------
        xy : numpy.ndarray
            receiver locations of shape (n_locations, 2)
        field : ("secondary", "total")
            Flag for the type of field to return.
        """
        sig = self.sigma_hat
        f = self.frequency
        w = 2*np.pi*f
        k = np.sqrt(-1j*w*mu_0*sig)[:, None]
        dxy = xy[:, :2] - self.location[:2]
        r = np.linalg.norm(dxy, axis=-1)
        x = dxy[:, 0]
        y = dxy[:, 1]

        em_x = em_y = em_z = 0
        src_x, src_y, src_z = self.orientation
        # Z component of source
        alpha = 1j*k*r/2.
        IK1 = iv(1, alpha)*kv(1, alpha)
        IK2 = iv(2, alpha)*kv(2, alpha)
        if src_z != 0.0:
            em_z += src_z*2.0/(k**2*r**5)*(9-(9+9*1j*k*r-4*k**2*r**2-1j*k**3*r**3)*np.exp(-1j*k*r))
            Hr = (k**2/r)*(IK1 - IK2)
            angle = np.arctan2(y, x)
            em_x += src_z*np.cos(angle)*Hr
            em_y += src_z*np.sin(angle)*Hr

        if src_x != 0.0 or src_y != 0.0:
            # X component of source
            phi = 2/(k**2*r**4)*(3 + k**2*r**2 - (3 + 3j*k*r - k**2*r**2)*np.exp(-1j*k*r))
            dphi_dr = 2/(k**2*r**5)*(-2*k**2*r**2 - 12 + (-1j*k**3*r**3 - 5*k**2*r**2 + 12j*k*r + 12)*np.exp(-1j*k*r))
            if src_x != 0.0:
                em_x += src_x*(-1.0/r**3)*(y**2*phi + x**2*r*dphi_dr)
                em_y += src_x*(1.0/r**3)*x*y*(phi - r*dphi_dr)
                em_z -= src_x*(k**2*x/r**2)*(IK1 - IK2)

            # Y component of source
            if src_y != 0.0:
                em_x += src_y*(1.0/r**3)*x*y*(phi - r*dphi_dr)
                em_y += src_y*(-1.0/r**3)*(x**2*phi + y**2*r*dphi_dr)
                em_z -= src_y*(k**2*y/r**2)*(IK1 - IK2)

        if field == "secondary":
            # subtract out primary field from above
            mdotr = src_x*x + src_y*y# + m[2]*(z=0)

            em_x -= 3*x*mdotr/r**5 - src_x/r**3
            em_y -= 3*y*mdotr/r**5 - src_y/r**3
            em_z -= -src_z/r**3 # + 3*(z=0)*mdotr/r**5

        return self.moment/(4*np.pi)*np.stack((em_x, em_y, em_z), axis=-1)



import numpy as np
from geoana.em.fdem.base import BaseFDEM
from geoana.spatial import repeat_scalar
from geoana.em.base import BaseElectricDipole, BaseMagneticDipole


class ElectricDipoleWholeSpace(BaseElectricDipole, BaseFDEM):
    """
    Harmonic electric dipole in a whole space. The source is
    (c.f. Ward and Hohmann, 1988 page 173). The source current
    density for a dipole located at :math:`\\mathbf{r}_s` with orientation
    :math:`\\mathbf{\\hat{u}}`

    .. math::

        \\mathbf{J}(\\mathbf{r}) = I ds \\delta(\\mathbf{r}
        - \\mathbf{r}_s)\\mathbf{\\hat{u}}

    """
    def vector_potential(self, xyz):
        """
        Vector potential for an electric dipole in a wholespace

        .. math::

            \\mathbf{A} = \\frac{I ds}{4 \\pi r} e^{-ikr}\\mathbf{\\hat{u}}

        """
        r = self.distance(xyz)
        a = (
            (self.current * self.length) / (4*np.pi*r) *
            np.exp(-1j*self.wavenumber*r)
        )
        a = np.kron(np.ones(1, 3), np.atleast_2d(a).T)
        return self.dot_orientation(a)

    def electric_field(self, xyz):
        """
        Electric field from an electric dipole

        .. math::

            \\mathbf{E} = \\frac{1}{\\hat{\\sigma}} \\nabla \\nabla \\cdot
             \\mathbf{A}
            - i \\omega \\mu \\mathbf{A}

        """
        dxyz = self.vector_distance(xyz)
        r = repeat_scalar(self.distance(xyz))
        kr = self.wavenumber * r
        ikr = 1j * kr

        front_term = (
            (self.current * self.length) / (4 * np.pi * self.sigma * r**3) *
            np.exp(-ikr)
        )
        symmetric_term = (
            repeat_scalar(self.dot_orientation(dxyz)) * dxyz *
            (-kr**2 + 3*ikr + 3) / r**2
        )
        oriented_term = (
            (kr**2 - ikr - 1) *
            np.kron(self.orientation, np.ones((dxyz.shape[0], 1)))
        )
        return front_term * (symmetric_term + oriented_term)

    def current_density(self, xyz):
        """
        Current density due to a harmonic electric dipole
        """
        return self.sigma * self.electric_field(xyz)

    def magnetic_field(self, xyz):
        """
        Magnetic field from an electric dipole

        .. math::

            \\mathbf{H} = \\nabla \\times \\mathbf{A}

        """
        dxyz = self.vector_distance(xyz)
        r = repeat_scalar(self.distance(xyz))
        kr = self.wavenumber * r
        ikr = 1j*kr

        front_term = (
            self.current * self.length / (4 * np.pi * r**2) * (ikr + 1) *
            np.exp(-ikr)
        )
        return -front_term * self.cross_orientation(dxyz) / r

    def magnetic_flux_density(self, xyz):
        """
        magnetic flux density from an electric dipole
        """
        return self.mu * self.magnetic_field(xyz)


class MagneticDipoleWholeSpace(BaseMagneticDipole, BaseFDEM):
    """
    Harmonic magnetic dipole in a whole space.
    """

    def vector_potential(self, xyz):
        """
        Vector potential for a magnetic dipole in a wholespace

        .. math::

            \\mathbf{F} = \\frac{i \\omega \\mu m}{4 \\pi r} e^{-ikr}
            \\mathbf{\\hat{u}}

        """
        r = self.distance(xyz)
        f = (
            (1j * self.omega * self.mu * self.moment) / (4 * np.pi * r) *
            np.exp(-1j * self.wavenumber * r)
        )
        f = np.kron(np.ones(1, 3), np.atleast_2d(f).T)
        return self.dot_orientation(f)

    def electric_field(self, xyz):
        """
        Electric field from a magnetic dipole in a wholespace
        """
        dxyz = self.vector_distance(xyz)
        r = repeat_scalar(self.distance(xyz))
        kr = self.wavenumber*r
        ikr = 1j * kr

        front_term = (
            (1j * self.omega * self.mu * self.moment) / (4. * np.pi * r**2) *
            (ikr + 1) * np.exp(-ikr)
        )
        return front_term * self.cross_orientation(dxyz) / r

    def current_density(self, xyz):
        """
        Current density from a magnetic dipole in a wholespace
        """
        return self.sigma * self.electric_field(xyz)

    def magnetic_field(self, xyz):
        """
        Magnetic field due to a magnetic dipole in a wholespace
        """
        dxyz = self.vector_distance(xyz)
        r = repeat_scalar(self.distance(xyz))
        kr = self.wavenumber*r
        ikr = 1j*kr

        front_term = self.moment / (4. * np.pi * r**3) * np.exp(-ikr)
        symmetric_term = (
            repeat_scalar(self.dot_orientation(dxyz)) * dxyz *
            (-kr**2 + 3*ikr + 3) / r**2
        )
        oriented_term = (
            (kr**2 - ikr - 1) *
            np.kron(self.orientation, np.ones((dxyz.shape[0], 1)))
        )

        return front_term * (symmetric_term + oriented_term)

    def magnetic_flux_density(self, xyz):
        """
        Magnetic flux density due to a magnetic dipole in a wholespace
        """
        return self.mu * self.magnetic_field(xyz)



from geoana.em.base import BaseEM
import numpy as np
from scipy.constants import mu_0, epsilon_0
import properties

###############################################################################
#                                                                             #
#                           Utility Functions                                 #
#                                                                             #
###############################################################################


def omega(frequency):
    """
    Angular frequency

    .. math::

        \\omega = 2 \\pi f

    **Required**
    :param frequency float: frequency (Hz)

    """
    return 2*np.pi*frequency


def wavenumber(
    frequency, sigma, mu=mu_0, epsilon=epsilon_0, quasistatic=False
):
    """
    Wavenumber of an electromagnetic wave in a medium with constant physical
    properties

    .. math::

        k = \\sqrt{\\omega^2 \\mu \\varepsilon - i \\omega \\mu \\sigma}


    **Required**

    :param (float, numpy.ndarray) frequency: frequency (Hz)
    :param float sigma: electrical conductivity (S/m)

    **Optional**

    :param float mu: magnetic permeability (H/m). Default: :math:`\\mu_0 = 4\\pi \\times 10^{-7}` H/m
    :param float epsilon: dielectric permittivity (F/m). Default: :math:`\\epsilon_0 = 8.85 \\times 10^{-12}` F/m
    :param bool quasistatic: use the quasi-static assumption? Default: False

    """
    w = omega(frequency)
    if quasistatic is True:
        return np.sqrt(-1j * w * mu * sigma)
    return np.sqrt(w**2 * mu * epsilon - 1j * w * mu * sigma)


def skin_depth(frequency, sigma, mu=mu_0):
    """
    Distance at which an em wave has decayed by a factor of :math:`1/e` in a
    medium with constant physical properties

    .. math::

        \\sqrt{\\frac{2}{\\omega \\sigma \\mu}}

    **Required**

    :param float frequency: frequency (Hz)
    :param float sigma: electrical conductivity (S/m)

    **Optional**
    :param float mu: magnetic permeability (H/m). Default: :math:`\mu_0 = 4\pi \\times 10^{-7}` H/m

    """
    w = omega(frequency)
    return np.sqrt(2./(w*sigma*mu))


def sigma_hat(frequency, sigma, epsilon=epsilon_0, quasistatic=False):
    """
    conductivity with displacement current contribution

    .. math::

        \hat{\sigma} = \sigma + i \omega \\varepsilon

    **Required**

    :param (float, numpy.array) frequency: frequency (Hz)
    :param float sigma: electrical conductivity (S/m)

    **Optional**

    :param float epsilon: dielectric permittivity. Default :math:`\\varepsilon_0`
    :param bool quasistatic: use the quasi-static assumption? Default: False

    """
    if quasistatic is True:
        return sigma
    return sigma + 1j*omega(frequency)*epsilon


###############################################################################
#                                                                             #
#                                  Classes                                    #
#                                                                             #
###############################################################################

class BaseFDEM(BaseEM):
    """
    Base frequency domain electromagnetic class
    """
    sigma = properties.Complex(
        "Electrical conductivity (S/m)",
        default=1.0,
        cast=True
    )

    frequency = properties.Float(
        "Source frequency (Hz)",
        default=1.,
        min=0.0
    )

    quasistatic = properties.Bool(
        "Use the quasi-static approximation and ignore displacement current?",
        default=False
    )

    @properties.validator('sigma')
    def _validate_real_part(self, change):
        if not np.real(change['value']) > 0:
            raise properties.ValidationError("The real part of sigma must be positive")

    @property
    def omega(self):
        """
        Angular frequency

        .. math::

            \\omega = 2\\pi f

        """
        return omega(self.frequency)

    @property
    def sigma_hat(self):
        """
        conductivity with displacement current contribution

        .. math::

            \\hat{\\sigma} = \\sigma + i \\omega \\varepsilon

        """
        sigma = sigma_hat(
            self.frequency, self.sigma, epsilon=self.epsilon,
            quasistatic=self.quasistatic
        )
        if np.all(np.imag(sigma) == 0):
            sigma = np.real(sigma)
        return sigma

    @property
    def wavenumber(self):
        """
        Wavenumber of an electromagnetic wave in a medium with constant
        physical properties

        .. math::

            k = \\sqrt{\\omega**2 \\mu \\varepsilon - i \\omega \\mu \\sigma}

        """
        return wavenumber(
            self.frequency, self.sigma, mu=self.mu, epsilon=self.epsilon,
            quasistatic=self.quasistatic
        )

    @property
    def skin_depth(self):
        """
        Distance at which an em wave has decayed by a factor of :math:`1/e` in
        a medium with constant physical properties

        .. math::

            \\sqrt{\\frac{2}{\\omega \\sigma \\mu}}

        """
        return skin_depth(self.frequency, self.sigma, mu=self.mu)
